<?php

/**
 * This files adds DivideBuy payment method in woocommerce. At admin side,
 *
 * go to woocommerce->settings->checkout where new method will be added.
 */

if ( ! defined( 'ABSPATH' ) ) { 
    exit; // Exit if file accessed directly
}

// Including our Gateway class and registering Payment Gateway with WooCommerce
add_action( 'plugins_loaded', 'init_divide_buy_gateway_class' );

function init_divide_buy_gateway_class() {

    // Adding dividebuy payment gatweway class
    class WC_Gateway_Divide_Buy extends WC_Payment_Gateway {

        /**
         * Initializing constructor for gateway
         */
        public function __construct() {

            $this->id                   = 'dividebuy';
            $this->has_fields           = false;
            $this->order_button_text    = __( 'Proceed to DivideBuy', 'woocommerce' );
            $this->method_title         = __( 'DivideBuy', 'woocommerce' );
            $this->method_description   = __( 'DivideBuy helps customers to buy products with easy instalments.', 'woocommerce' );
            $this->supports             = array( 'products', 'refunds' );

            // Loading all gateway settings.
            $this->init_form_fields();
            $this->init_settings();

            // Defining user set variables.
            $this->title        = $this->get_option( 'title' );
            $this->description  = $this->get_option( 'description' );

            // Adding action for saving payment method details.
            add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
        }
                                
        /**
         * Initializing Gateway Settings Form Fields
         */
        public function init_form_fields() {
                        
                        $countries_obj   = new WC_Countries();
                        $countries   = $countries_obj->__get('countries');
                                                                        
            $this->form_fields = apply_filters( 'wc_divide_buy_form_fields', array(

                    'enabled' => array(
                        'title'   => __( 'Enable/Disable', 'woocommerce' ),
                        'type'    => 'checkbox',
                        'label'   => __( 'Enable DivideBuy Payment', 'woocommerce' ),
                        'default' => 'yes'
                    ),

                    'title' => array(
                        'title'       => __( 'Title', 'woocommerce' ),
                        'type'        => 'text',
                        'description' => __( 'This controls the title for the payment method the customer sees during checkout.', 'woocommerce' ),
                        'default'     => __( 'Divide Buy', 'woocommerce' ),
                        'desc_tip'    => true,
                    ),

                    'description' => array(
                        'title'       => __( 'Description', 'woocommerce' ),
                        'type'        => 'textarea',
                        'description' => __( 'Payment method description that the customer will see on checkout.', 'woocommerce' ),
                        'default'     => __( 'Please select products which are divide buy supported.', 'woocommerce' ),
                        'desc_tip'    => true,
                    ),
                
                    'gateway_banner_url' => array(
                        'title'       => __( 'DivideBuy Banner', 'woocommerce' ),
                        'type'        => 'text',
                        'description' => __( 'Enter URL for DivideBuy banner which will get displayed in checkout page.', 'woocommerce' ),
                        'desc_tip'    => true,
                    ),
            ) );
        }
                
                // Placing new order when clicked on checkout button.
        public function process_payment( $order_id ) {
                    
            global $woocommerce;
            $order = new WC_Order( $order_id );
            $postcodes = get_option('wc_settings_tab_dividebuy_checkout_postcode_list');
            $dividebuy_postcodes = explode(",", $postcodes);
            $is_deliverable = "true";
            
            if(trim($woocommerce->customer->get_shipping_postcode()) != ""){
                $shipping_post_code = trim($woocommerce->customer->get_shipping_postcode());
            }else{
                $shipping_post_code = trim($woocommerce->customer->get_billing_postcode());
            }
            
            foreach($dividebuy_postcodes as $code){
                if(trim($code) == $shipping_post_code){
                    $is_deliverable = "false";
                    break;
                }
            }

            // Getting order minimum value
            $dividebuy_helper = new DivideBuy_helper();
            $order_minimum_value = $dividebuy_helper->check_minimum_order_amount();
            $non_dividebuy_products_count = $dividebuy_helper->get_nondividebuy_products_count();
            
            if($non_dividebuy_products_count > 0){
                wp_delete_post( $order_id, true );
                throw new Exception(__("Error in processing checkout. There are non DivideBuy products in cart."));
            }
            else if($is_deliverable == "false"){
                wp_delete_post( $order_id, true );
                throw new Exception(__("Unfortunately we are unable to deliver to this postcode. Please call our Customer Service team for more details: 0800 085 0885."));
            }                        
            else if( $order_minimum_value['status'] == "false" ){
                wp_delete_post( $order_id, true );
                throw new Exception( __( 'Minimum order product total should be '. get_woocommerce_currency_symbol().$order_minimum_value['min_order_amount'] . '.' ,'woocommerce' ) );
            } else {
                
                $address = array(
                    'first_name' => $order->billing_first_name,
                    'last_name'  => $order->billing_last_name,
                    'email'      => $order->billing_email,
                    'phone'      => $order->billing_phone,
                    'address_1'  => $order->billing_address_1,
                    'address_2'  => $order->billing_address_2,
                    'city'       => $order->billing_city,
                    'county'     => $order->billing_state,
                    'postcode'   => $order->billing_postcode,
                    'country'    => $order->billing_country
                );
                
                $order->set_address( $address, 'billing' );
                $order->set_address( $address, 'shipping' );
                
                // Hide order details from retailer while payment is being confirmed.
                $order->update_status('freeze-dividebuy', __( 'Waiting for confirmation from DivideBuy.', 'woocommerce' ));
                update_post_meta( $order->id, 'order_is_dividebuy', 'yes' );
            
                $url = Divide_Buy_Configuration::get_order_redirect_url();
                $splashKey = base64_encode($order_id.':'.get_option('wc_settings_tab_retailer_id'));
                $redirect_url = $url.'?splashKey='.$splashKey;

                $woocommerce->cart->empty_cart();
                $dividebuy_helper->add_session_products_to_cart();

                // Redirecting to thank you page.
                return array(
                        'result'    => 'success',
                        'redirect'  => $redirect_url
                );
            }
        }

        public function process_refund($order_id, $amount = null, $reason = '')
        {

            return true;
        }
    }

        // Including DivideBuy payment with all available gateways.
    function wc_divide_buy_add_to_gateway( $gateways ) {
                
        $gateways[] = 'WC_Gateway_Divide_Buy';
        return $gateways;
    }
    add_filter( 'woocommerce_payment_gateways', 'wc_divide_buy_add_to_gateway' );
        
        // Adding DivideBuy payment gateway icon.
        function dividebuy_gateway_icon( $icon, $id ) {
            
                $gateway_class_obj = new WC_Gateway_Divide_Buy;
                $gateway_icon = $gateway_class_obj->get_option( 'gateway_banner_url' );
                
                if ( $id === 'dividebuy' ) {
                    if( $gateway_icon !='' ) {
                        return '<img src="' . $gateway_icon . '" alt="Divide Buy" />'; 
                    } else {
                        return '<img src="' . plugins_url(). '/dividebuy/assets/images/dividebuy-logo.png' . '" alt="Divide Buy" />'; 
                    }                    
                } else {
                    return $icon;
                }
        }
        add_filter( 'woocommerce_gateway_icon', 'dividebuy_gateway_icon', 10, 2 );

}

// Processing order refund
add_action( 'woocommerce_order_refunded', 'action_woocommerce_order_refunded', 10, 2 ); 
function action_woocommerce_order_refunded( $order_get_id, $refund_get_id ) { 
    
    global $wpdb, $woocommerce;
    $sql_reports_refunded = "
          SELECT
            product_id.meta_value        AS product_id,
            variation_id.meta_value      AS variation_id,
            order_items.order_item_name  AS product_name,
            product_sku.meta_value       AS product_SKU,
            variation_sku.meta_value     AS variation_SKU,
            SUM(quantity.meta_value)     AS quantity,
            line_subtotal_tax.meta_value     AS line_tax

          FROM {$wpdb->prefix}posts AS posts
          LEFT JOIN {$wpdb->prefix}woocommerce_order_items     AS order_items    ON posts.ID = order_items.order_id
          LEFT JOIN {$wpdb->prefix}woocommerce_order_itemmeta  AS quantity       ON order_items.order_item_id = quantity.order_item_id      AND quantity.meta_key      = '_qty'
          LEFT JOIN {$wpdb->prefix}woocommerce_order_itemmeta  AS line_total       ON order_items.order_item_id = line_total.order_item_id      AND line_total.meta_key      = '_line_total'
          LEFT JOIN {$wpdb->prefix}woocommerce_order_itemmeta  AS line_subtotal_tax       ON order_items.order_item_id = line_subtotal_tax.order_item_id     AND line_subtotal_tax.meta_key      = '_line_subtotal_tax'
          LEFT JOIN {$wpdb->prefix}woocommerce_order_itemmeta  AS variation_id   ON order_items.order_item_id = variation_id.order_item_id  AND variation_id.meta_key  = '_variation_id'
          LEFT JOIN {$wpdb->prefix}woocommerce_order_itemmeta  AS product_id     ON order_items.order_item_id = product_id.order_item_id    AND product_id.meta_key    = '_product_id'
          LEFT JOIN {$wpdb->prefix}postmeta                    AS product_sku    ON product_id.meta_value     = product_sku.post_id         AND product_sku.meta_key   = '_sku'
          LEFT JOIN {$wpdb->prefix}postmeta                    AS variation_sku  ON variation_id.meta_value   = variation_sku.post_id       AND variation_sku.meta_key = '_sku'
          WHERE  
          CASE WHEN posts.post_type = 'shop_order' THEN 
                 posts.post_type IN ('shop_order') AND posts.post_status = 'wc-refunded' 
            ELSE
                posts.post_type IN ('shop_order_refund')
            END
            AND posts.ID = '{$order_get_id}'
            AND order_items.order_item_type = 'line_item'
            AND order_items.order_id IS NOT NULL
          GROUP BY product_id, variation_id
          ORDER BY quantity DESC, product_name ASC";

    $orderTotal = get_post_meta($order_get_id,'_order_total',true);
    $data_refund = $wpdb->get_results($sql_reports_refunded,ARRAY_A);
    $data = array();
    if(count($data_refund) > 0) {   
        
        $data['product'] = array();
        $i = 0;

        foreach($data_refund as $data_refund) {
            if($data_refund['variation_sku'] != '') {
                $data['product'][$i]['sku'] = $data_refund['variation_sku'];    
            }elseif($data_refund['product_sku'] != '') {
                $data['product'][$i]['sku'] = $data_refund['product_sku'];    
            }else {
                $data['product'][$i]['sku'] = $data_refund['product_id'];    
            }

            $data['product'][$i]['productName'] = $data_refund['product_name'];
            $data['product'][$i]['qty'] = str_replace("-", "", $data_refund['quantity']);
            if($data_refund['variation_id'] != '0') {   
                $_product = wc_get_product( $data_refund['variation_id'] );
                $Price = $_product->get_price();    
            }else {
                $_product = wc_get_product( $data_refund['product_id'] );
                $Price = $_product->get_price();    
            }            

            $data['product'][$i]['rowTotal'] = trim($Price);
            
            $data['product'][$i]['rowInclTotal'] =  str_replace("-", "",$data_refund['line_tax']);
            $i++;    
        }
    }

    $data['product'] = $data['product'];
    $data['totalRefund'] = get_post_meta($refund_get_id,'_refund_amount',true);
    $data['orderId'] = $order_get_id;
    $data['reason'] = get_post_meta($refund_get_id,'_refund_reason',true);
    $data['shippingCostAmount'] = str_replace("-", "",get_post_meta($refund_get_id,'_order_shipping',true));
    $data['shippingTaxAmount'] = str_replace("-", "",get_post_meta($refund_get_id,'_order_shipping_tax',true));
    $data['adjustmentRefund'] = "";
    $data['adjustmentFee'] = "";
    $data['taxAmount'] = "";
    $data['discountAmount'] = str_replace("-", "",get_post_meta($refund_get_id,'_cart_discount',true));
    $totalRefundedAmount = get_post_meta($refund_get_id,'_refund_amount',true);

    $orderTotal = trim($orderTotal);
    $orderTotal = ltrim($orderTotal, "-");
    
    if( $orderTotal == $totalRefundedAmount ) {
        $data['refundType'] = 'FULL';
    }else {
        $data['refundType'] = 'PARTIAL';
    }

    $data['retailer'] = array();
    $data['retailer']['retailerId']= get_option('wc_settings_tab_retailer_id');
    $data['retailer']['storeAuthentication'] = get_option('wc_settings_tab_retailer_authentication_number');
    $data['retailer']['storeToken'] = get_option('wc_settings_tab_retailer_token_number');
    
    $data_json = json_encode($data);
    $url = Divide_Buy_Configuration::get_api_url().'api/refund';
    $method = 'POST';
    $send_refund_data = DivideBuy_Api::curl_call_api($url,$data_json,$method);
    
    return true;
}